import time
import os
from threading import Thread
import socket
import subprocess
import shutil

from terminaltables import AsciiTable

from ltcli import config
from ltcli.log import logger
from ltcli import net
from ltcli.rediscli_util import RedisCliUtil
from ltcli.redistrib2 import command as trib
from ltcli import utils
from ltcli.deploy_util import DeployUtil, DEPLOYED
from ltcli import ask_util
from ltcli import color
from ltcli.exceptions import (
    SSHConnectionError,
    HostConnectionError,
    HostNameError,
    ClusterRedisError,
    ClusterNotExistError,
)


def get_ps_list_command(port_list):
    port_filter = '|'.join(str(x) for x in port_list)
    command = [
        "ps -ef",
        "grep 'redis-server'",
        "egrep '({})'".format(port_filter),
        "grep -v 'ps -ef'",
        "grep -v 'grep'",
    ]
    command = ' | '.join(command)
    return command


def get_my_ps_list_command(port_list, cluster_id):
    port_filter = '|'.join(str(x) for x in port_list)
    command = [
        "ps -ef",
        "grep 'redis-server'",
        "grep `whoami`",
        "grep cluster_{}".format(cluster_id),
        "egrep '({})'".format(port_filter),
        "grep -v 'ps -ef'",
        "grep -v 'grep'",
    ]
    command = ' | '.join(command)
    return command


class Center(object):
    def __init__(self):
        self.master_host_list = []
        self.slave_host_list = []
        self.master_port_list = []
        self.slave_port_list = []
        self.all_host_list = []

    def sync_conf(self, show_result=False):
        logger.info('sync conf')
        path_of_fb = config.get_path_of_fb(self.cluster_id)
        conf_path = path_of_fb['conf_path']
        my_address = config.get_local_ip_list()
        meta = [['HOST', 'STATUS']]
        no_cluster_flag = False
        for host in self.all_host_list:
            client = net.get_ssh(host)
            cluster_path = path_of_fb['cluster_path']
            if not net.is_dir(client, cluster_path):
                no_cluster_flag = True
                meta.append([host, color.red('NO CLUSTER')])
                continue
            client.close()
            meta.append([host, color.green('OK')])
        if no_cluster_flag:
            utils.print_table(meta)
            logger.error('Cancel sync conf')
            return False
        meta = [['HOST', 'STATUS']]
        error_flag = False
        for host in self.all_host_list:
            if net.get_ip(host) in my_address:
                meta.append([host, color.green('OK')])
                continue
            client = net.get_ssh(host)
            try:
                net.copy_dir_to_remote(client, conf_path, conf_path)
                meta.append([host, color.green('OK')])
            except BaseException as ex:
                logger.debug(ex)
                meta.append([host, color.red('FAIL')])
                error_flag = True
            finally:
                client.close()
        if error_flag or show_result:
            utils.print_table(meta)
        if error_flag:
            logger.error('Fail conf sync')
            return False
        logger.info('OK')
        return True

    def sync_file(self, file_path, show_result=False):
        logger.info('sync conf')
        meta = [['HOST', 'STATUS']]
        my_address = config.get_local_ip_list()
        error_flag = False
        for host in self.all_host_list:
            client = net.get_ssh(host)
            sftp = net.get_sftp(client)
            if net.get_ip(host) in my_address:
                meta.append([host, color.green('OK')])
                continue
            try:
                sftp.put(file_path, file_path)
                meta.append([host, color.green('OK')])
            except Exception as ex:
                logger.debug(ex)
                meta.append([host, color.red('FAIL')])
                error_flag = True
            finally:
                sftp.close()
                client.close()
        if error_flag or show_result:
            utils.print_table(meta)
        if error_flag:
            logger.error('Fail file sync')
            return False
        logger.info('OK')
        return True

    def configure_redis(self, master=True, slave=True):
        logger.debug('configure redis')
        path_of_fb = config.get_path_of_fb(self.cluster_id)
        sr2_redis_conf_temp = path_of_fb['sr2_redis_conf_temp']
        if os.path.exists(sr2_redis_conf_temp):
            shutil.rmtree(sr2_redis_conf_temp)
        os.mkdir(sr2_redis_conf_temp)
        sr2_redis_conf = path_of_fb['sr2_redis_conf']
        if not os.path.exists(sr2_redis_conf):
            os.mkdir(sr2_redis_conf)
        props_path = path_of_fb['redis_properties']
        prefix_srd = config.get_props(props_path, 'sr2_redis_data')
        prefix_sfdp = config.get_props(props_path, 'sr2_flash_db_path')
        ssd_count = config.get_props(props_path, 'ssd_count')
        sr2_redis_conf_temp = path_of_fb['sr2_redis_conf_temp']
        conf_path = path_of_fb['conf_path']
        user = os.environ['USER']
        command = []
        count = 0
        if master:
            template = 'redis-master.conf.template'
            template_path = os.path.join(conf_path, template)
            for port in self.master_port_list:
                # depence: [Errno 7] Argument list too long
                if count > 200:
                    command = ' '.join(command)
                    subprocess.check_output(command, shell=True)
                    logger.debug('subprocess: {}'.format(command))
                    command = []
                    count = 0
                count += 1
                ssd_no = config.get_sata_ssd_no(port, ssd_count)
                sr2_redis_data = '{}{}/{}'.format(prefix_srd, ssd_no, user)
                sr2_flash_db_path = '{}{}/{}/db/db-{}'.format(
                    prefix_sfdp,
                    ssd_no,
                    user,
                    port
                )
                file_name = 'redis-{}.conf'.format(port)
                export_envs = ' '.join([
                    'export SR2_REDIS_PORT={}'.format(port),
                    'export SR2_REDIS_DATA={}'.format(sr2_redis_data),
                    'export SR2_FLASH_DB_PATH={}'.format(sr2_flash_db_path),
                ])
                target = os.path.join(sr2_redis_conf_temp, file_name)
                command.append('{}; cat {} | envsubst > {};'.format(
                    export_envs,
                    template_path,
                    target,
                ))
        if slave and self.slave_port_list:
            template = 'redis-slave.conf.template'
            template_path = os.path.join(conf_path, template)
            for port in self.slave_port_list:
                # depence: [Errno 7] Argument list too long
                if count > 200:
                    command = ' '.join(command)
                    subprocess.check_output(command, shell=True)
                    logger.debug('subprocess: {}'.format(command))
                    command = []
                    count = 0
                count += 1
                ssd_no = config.get_sata_ssd_no(port, ssd_count)
                sr2_redis_data = '{}{}/{}'.format(prefix_srd, ssd_no, user)
                sr2_flash_db_path = '{}{}/{}/db/db-{}'.format(
                    prefix_sfdp,
                    ssd_no,
                    user,
                    port
                )
                file_name = 'redis-{}.conf'.format(port)
                export_envs = ' '.join([
                    'export SR2_REDIS_PORT={}'.format(port),
                    'export SR2_REDIS_DATA={}'.format(sr2_redis_data),
                    'export SR2_FLASH_DB_PATH={}'.format(sr2_flash_db_path),
                ])
                target = os.path.join(sr2_redis_conf_temp, file_name)
                command.append('{}; cat {} | envsubst > {};'.format(
                    export_envs,
                    template_path,
                    target,
                ))
        command = ' '.join(command)
        subprocess.check_output(command, shell=True)
        logger.debug('subprocess: {}'.format(command))
        command = 'find {} -type f | xargs -i cp "{{}}" {}'.format(
            sr2_redis_conf_temp,
            sr2_redis_conf
        )
        subprocess.check_output(command, shell=True)
        logger.debug('subprocess: {}'.format(command))

    def backup_server_logs(self, master=True, slave=True):
        """ Backup server logs

        Move redis log files to $SR2_REDIS_LOG/backup/<time-stamp>
        """
        logger.debug('backup_server_logs')
        current_time = time.strftime("%Y%m%d-%H%M%S", time.gmtime())
        path_of_fb = config.get_path_of_fb(self.cluster_id)
        sr2_redis_log = path_of_fb['sr2_redis_log']
        backup_path = os.path.join(sr2_redis_log, 'backup', current_time)
        logger.debug('backup path: {}'.format(backup_path))
        if master:
            logger.info('Backup redis master log in each MASTER hosts...')
            self._backup_server_logs(
                self.master_host_list,
                self.master_port_list,
                backup_path
            )
        if slave and self.slave_host_list:
            logger.info('Backup redis slave log in each SLAVE hosts...')
            self._backup_server_logs(
                self.slave_host_list,
                self.slave_port_list,
                backup_path
            )

    def _backup_server_logs(self, hosts, ports, backup_path):
        path_of_fb = config.get_path_of_fb(self.cluster_id)
        sr2_redis_log = path_of_fb['sr2_redis_log']
        for host in hosts:
            logger.info(' - {}'.format(host))
            command = ['mkdir -p {};'.format(backup_path)]
            count = 0
            for port in ports:
                # depence: [Errno 7] Argument list too long
                if count > 100:
                    command = ' '.join(command)
                    client = net.get_ssh(host)
                    net.ssh_execute(client, command, allow_status=[0, 1])
                    client.close()
                    command = []
                    count = 0
                count += 1
                command.append('mv {0}/*{1}.log {2} &> /dev/null'.format(
                    sr2_redis_log,
                    port,
                    backup_path
                ))
            command = ' '.join(command)
            client = net.get_ssh(host)
            net.ssh_execute(client, command, allow_status=[0, 1])
            client.close()

    def conf_backup(self, host, cluster_id, tag):
        logger.info('Backup conf of cluster {}...'.format(cluster_id))
        # prepare
        path_of_fb = config.get_path_of_fb(cluster_id)
        conf_path = path_of_fb['conf_path']
        path_of_cli = config.get_path_of_cli(cluster_id)
        conf_backup_path = path_of_cli['conf_backup_path']
        conf_backup_tag_path = os.path.join(conf_backup_path, tag)

        if not os.path.isdir(conf_backup_path):
            os.mkdir(conf_backup_path)

        # back up conf
        os.mkdir(conf_backup_tag_path)
        client = net.get_ssh(host)
        net.copy_dir_from_remote(client, conf_path, conf_backup_tag_path)
        client.close()

        logger.info('OK, {}'.format(tag))

    def cluster_backup(self, host, cluster_id, tag):
        logger.info('Backup cluster {} at {}...'.format(cluster_id, host))
        # prepare
        path_of_fb = config.get_path_of_fb(cluster_id)
        cluster_path = path_of_fb['cluster_path']
        cluster_backup_path = path_of_fb['cluster_backup_path']
        cluster_backup_tag_path = os.path.join(cluster_backup_path, tag)

        # back up cluster
        client = net.get_ssh(host)
        if not net.is_dir(client, cluster_backup_path):
            sftp = net.get_sftp(client)
            sftp.mkdir(cluster_backup_path)
            sftp.close()
        command = 'mv {} {}'.format(cluster_path, cluster_backup_tag_path)
        net.ssh_execute(client=client, command=command)
        client.close()
        logger.info('OK, {}'.format(tag))

    def conf_restore(self, host, cluster_id, tag):
        logger.info('Restore conf to cluster {}...'.format(cluster_id))
        # prepare
        path_of_fb = config.get_path_of_fb(cluster_id)
        path_of_cli = config.get_path_of_cli(cluster_id)
        conf_path = path_of_fb['conf_path']
        conf_backup_path = path_of_cli['conf_backup_path']
        conf_backup_tag_path = os.path.join(conf_backup_path, tag)

        # restore conf
        client = net.get_ssh(host)
        net.copy_dir_to_remote(client, conf_backup_tag_path, conf_path)
        client.close()
        logger.debug('OK')

    def get_alive_redis_count(self, hosts, ports, check_owner=False):
        logger.debug('get_alive_redis_count')
        logger.debug('hosts={}, ports={}'.format(hosts, ports))
        if check_owner:
            ps_list_command = get_my_ps_list_command(ports, self.cluster_id)
        else:
            ps_list_command = get_ps_list_command(ports)
        command = '{} | wc -l'.format(ps_list_command)
        total = 0
        for host in hosts:
            client = net.get_ssh(host)
            _, stdout_msg, _ = net.ssh_execute(client=client, command=command)
            total += int(stdout_msg.strip())
        logger.debug('redis-server total={}'.format(total))
        cmd = "ps -ef | grep 'redis-rdb-to-slaves' | grep -v 'grep' | wc -l"
        redis_rdb_count = 0
        for host in hosts:
            client = net.get_ssh(host)
            _, stdout_msg, _ = net.ssh_execute(client=client, command=cmd)
            redis_rdb_count += int(stdout_msg.strip())
        logger.debug(command)
        logger.debug('redis-rbd-to-slaves total={}'.format(total))
        total += redis_rdb_count
        return total

    def get_alive_master_redis_count(self, check_owner=False):
        logger.debug('get_alive_master_redis_count')
        hosts = self.master_host_list
        ports = self.master_port_list
        alive_count = self.get_alive_redis_count(hosts, ports, check_owner)
        logger.debug('alive master count={}'.format(alive_count))
        return alive_count

    def get_alive_slave_redis_count(self, check_owner=False):
        logger.debug('get_alive_slave_redis_count')
        hosts = self.slave_host_list
        ports = self.slave_port_list
        alive_count = self.get_alive_redis_count(hosts, ports, check_owner)
        logger.debug('alive slave count={}'.format(alive_count))
        return alive_count

    def get_alive_all_redis_count(self, check_owner=False):
        logger.debug('get_alive_all_redis_count')
        total_m = self.get_alive_master_redis_count(check_owner)
        total_s = self.get_alive_slave_redis_count(check_owner)
        return total_m + total_s

    def create_cluster(self, yes=False):
        """Create cluster
        """
        logger.info('>>> Creating cluster')
        logger.debug('create cluster start')
        result = self.confirm_node_port_info(skip=yes)
        if not result:
            logger.warn('Cancel create')
            return
        m_ip_list = list(map(net.get_ip, self.master_host_list))
        targets = utils.get_ip_port_tuple_list(
            m_ip_list,
            self.master_port_list
        )
        try:
            trib.create(targets, max_slots=16384)
        except Exception as ex:
            logger.error(str(ex))
            return
        if self.slave_port_list:
            self.replicate()
        logger.info('create cluster complete.')

    def confirm_node_port_info(self, skip=False):
        replicas = config.get_replicas(self.cluster_id)
        meta = [['HOST', 'PORT', 'TYPE']]
        for node in self.master_host_list:
            for port in self.master_port_list:
                meta.append([node, port, 'MASTER'])
        for node in self.slave_host_list:
            for port in self.slave_port_list:
                meta.append([node, port, 'SLAVE'])
        table = AsciiTable(meta)
        print(table.table)
        print('replicas: {}\n'.format(replicas))
        if skip:
            return True
        msg = [
            'Do you want to proceed with the replicate ',
            'according to the above information?',
        ]
        yes = ask_util.askBool(''.join(msg), ['y', 'n'])
        return yes

    def stop_redis_process(self, host, ports, force=False):
        """Stop redis process
        """
        logger.debug('stop_redis_process')
        signal = 'SIGKILL' if force else 'SIGINT'
        ps_list_command = get_ps_list_command(ports)
        pid_list = "{} | awk '{{print $2}}'".format(ps_list_command)
        command = 'kill -s {} $({})'.format(signal, pid_list)
        client = net.get_ssh(host)
        net.ssh_execute(client, command, allow_status=[-1, 0, 1, 2, 123, 130])
        client.close()

    def stop_redis(self, force=False, master=True, slave=True):
        """Stop redis

        :param force: If true, send SIGKILL. If not, send SIGINT
        """
        logger.debug('stop_redis')
        success = False
        if not self.slave_host_list:
            slave = False
        if slave:
            logger.info('Stopping slave cluster of redis...')
            s_ports = self.slave_port_list
            s_count = len(self.slave_host_list) * len(s_ports)
            max_try_count = 10
            while max_try_count > 0:
                alive_count = self.get_alive_slave_redis_count()
                logger.info('cur: {} / total: {}'.format(alive_count, s_count))
                if alive_count <= 0:
                    logger.info('Complete all slave redis process down')
                    success = True
                    break
                max_try_count -= 1
                if max_try_count % 3 == 0:
                    for host in self.slave_host_list:
                        self.stop_redis_process(host, s_ports, force)
                time.sleep(1)
        if slave and not success:
            raise ClusterRedisError('Fail to stop redis: max try exceed')
        success = False
        if master:
            logger.info('Stopping master cluster of redis...')
            m_ports = self.master_port_list
            m_count = len(self.master_host_list) * len(m_ports)
            max_try_count = 10
            while max_try_count > 0:
                alive_count = self.get_alive_master_redis_count()
                logger.info('cur: {} / total: {}'.format(alive_count, m_count))
                if alive_count <= 0:
                    logger.info('Complete all master redis process down')
                    success = True
                    break
                max_try_count -= 1
                if max_try_count % 3 == 0:
                    for host in self.master_host_list:
                        self.stop_redis_process(host, m_ports, force)
                time.sleep(1)
        if master and not success:
            raise ClusterRedisError('Fail to stop redis: max try exceed')

    def create_redis_data_directory(self, master=True, slave=True):
        """ create directory SR2_REDIS_DATA, SR2_FLASH_DB_PATH
        """
        logger.debug('create_redis_data_directory')
        if master:
            logger.info('create redis data directory in each MASTER hosts')
            self._create_redis_data_directory(
                self.master_host_list,
                self.master_port_list,
            )
        if slave and self.slave_host_list:
            logger.info('create redis data directory in each SLAVE hosts')
            self._create_redis_data_directory(
                self.slave_host_list,
                self.slave_port_list,
            )

    def _create_redis_data_directory(self, hosts, ports):
        path_of_fb = config.get_path_of_fb(self.cluster_id)
        props_path = path_of_fb['redis_properties']
        prefix_srd = config.get_props(props_path, 'sr2_redis_data')
        prefix_sfdp = config.get_props(props_path, 'sr2_flash_db_path')
        ssd_count = config.get_props(props_path, 'ssd_count')
        user = os.environ['USER']
        for host in hosts:
            logger.info(' - {}'.format(host))
            command = ['mkdir -p']
            count = 0
            for port in ports:
                # depence: [Errno 7] Argument list too long
                if count > 100:
                    command = ' '.join(command)
                    client = net.get_ssh(host)
                    net.ssh_execute(client, command)
                    client.close()
                    command = ['mkdir -p']
                    count = 0
                count += 1
                ssd_no = config.get_sata_ssd_no(port, ssd_count)
                sr2_redis_data = '{}{}/{}'.format(
                    prefix_srd,
                    ssd_no,
                    user
                )
                sr2_flash_db_path = '{}{}/{}/db/db-{}'.format(
                    prefix_sfdp,
                    ssd_no,
                    user,
                    port
                )
                command.append(sr2_redis_data)
                command.append(sr2_flash_db_path)
            command = ' '.join(command)
            client = net.get_ssh(host)
            net.ssh_execute(client, command)
            client.close()

    def wait_until_all_redis_process_up(self, master=True, slave=True):
        """Wait until all redis process up
        """
        logger.debug('wait_until_all_redis_process_up')
        logger.info('Wait until all redis process up...')
        total_count = 0
        if master:
            total_count += len(self.master_host_list) * len(self.master_port_list)
        if slave:
            total_count += len(self.slave_host_list) * len(self.slave_port_list)
        max_try_count = 10
        while max_try_count > 0:
            alive_count = 0
            if master:
                alive_count += self.get_alive_master_redis_count()
            if slave:
                alive_count += self.get_alive_slave_redis_count()
            logger.info('cur: {} / total: {}'.format(alive_count, total_count))
            if alive_count >= total_count:
                logger.info('Complete all redis process up')
                if alive_count != total_count:
                    logger.warning('ClusterRedisWarning: too many process up')
                return True
            time.sleep(1)
            max_try_count -= 1
            msg = [
                'Fail to start redis: max try exceed',
                "Recommendation Command: 'monitor'"
            ]
        # raise ClusterRedisError('Fail to start redis: max try exceed')
        raise ClusterRedisError('\n'.join(msg))

    def check_hosts_connection(self, hosts=None, show_result=False):
        logger.debug('check hosts connection')
        logger.info('Check status of hosts...')
        if hosts is None:
            self.update_ip_port()
            hosts = self.all_host_list
        host_status = []
        success_count = 0
        for host in hosts:
            try:
                client = net.get_ssh(host)
                client.close()
                logger.debug('{} ssh... OK'.format(host))
                success_count += 1
                host_status.append([host, color.green('OK')])
            except HostNameError:
                show_result = True
                host_status.append([host, color.red('UNKNOWN HOST')])
                logger.debug('{} gethostbyname... FAIL'.format(host))
            except HostConnectionError:
                show_result = True
                host_status.append([host, color.red('CONNECTION FAIL')])
                logger.debug('{} connection... FAIL'.format(host))
            except SSHConnectionError:
                show_result = True
                host_status.append([host, color.red('SSH FAIL')])
                logger.debug('{} ssh... FAIL'.format(host))
        if show_result:
            table = AsciiTable([['HOST', 'STATUS']] + host_status)
            print(table.table)
        if len(hosts) != success_count:
            return False
        logger.info('OK')
        return True

    def check_include_localhost(self, hosts):
        logger.debug('check_include_localhost')
        for host in hosts:
            try:
                ip_addr = socket.gethostbyname(host)
                if ip_addr in [config.get_local_ip(), '127.0.0.1']:
                    return True
            except socket.gaierror:
                raise HostNameError(host)
        return False

    def remove_all_of_redis_log_force(self):
        logger.info('remove all of redis log')
        path_of_fb = config.get_path_of_fb(self.cluster_id)
        sr2_redis_log = path_of_fb['sr2_redis_log']
        command = 'rm -f {}/*.log'.format(sr2_redis_log)
        for host in self.master_host_list:
            client = net.get_ssh(host)
            net.ssh_execute(client, command)
            client.close()
            logger.info(' - {}'.format(host))

    def remove_generated_config(self, client, port_list):
        path_of_fb = config.get_path_of_fb(self.cluster_id)
        sr2_redis_conf = path_of_fb['sr2_redis_conf']
        command = [
            'export SR2_REDIS_CONF={};'.format(sr2_redis_conf),
            'rm -rf'
        ]
        count = 0
        for port in port_list:
            if count > 100:
                command = ' '.join(command)
                net.ssh_execute(client, command)
                command = [
                    'export SR2_REDIS_CONF={};'.format(sr2_redis_conf),
                    'rm -rf'
                ]
                count = 0
            count += 1
            conf_file = 'redis-{}.conf'.format(port)
            conf_file_path = os.path.join('$SR2_REDIS_CONF', conf_file)
            command.append(conf_file_path)
        command = ' '.join(command)
        net.ssh_execute(client, command)

    def _remove_data(self, client, port_list):
        path_of_fb = config.get_path_of_fb(self.cluster_id)
        props_path = path_of_fb['redis_properties']
        prefix_srd = config.get_props(props_path, 'sr2_redis_data')
        prefix_sfdp = config.get_props(props_path, 'sr2_flash_db_path')
        ssd_count = config.get_props(props_path, 'ssd_count')
        user = os.environ['USER']
        command = ['rm -rf']
        count = 0
        for port in port_list:
            if count > 100:
                command = ' '.join(command)
                net.ssh_execute(client, command)
                command = ['rm -rf']
                count = 0
            count += 1
            ssd_no = config.get_sata_ssd_no(port, ssd_count)
            sr2_redis_data = '{}{}/{}'.format(
                prefix_srd,
                ssd_no,
                user
            )
            sr2_flash_db_path = '{}{}/{}/db/db-{}'.format(
                prefix_sfdp,
                ssd_no,
                user,
                port
            )
            command.append(sr2_flash_db_path)
            command.append('{}/appendonly-{}*.aof'.format(
                sr2_redis_data,
                port
            ))
            command.append('{}/dump-{}.rdb'.format(
                sr2_redis_data,
                port
            ))
        command = ' '.join(command)
        net.ssh_execute(client, command)

    def _remove_node_conf(self, client, port_list):
        path_of_fb = config.get_path_of_fb(self.cluster_id)
        props_path = path_of_fb['redis_properties']
        prefix_srd = config.get_props(props_path, 'sr2_redis_data')
        ssd_count = config.get_props(props_path, 'ssd_count')
        user = os.environ['USER']
        command = ['rm -f']
        count = 0
        for port in port_list:
            if count > 100:
                command = ' '.join(command)
                net.ssh_execute(client, command)
                command = ['rm -f']
                count = 0
            count += 1
            ssd_no = config.get_sata_ssd_no(port, ssd_count)
            sr2_redis_data = '{}{}/{}'.format(prefix_srd, ssd_no, user)
            file_name = 'nodes-{}.conf'.format(port)
            file_path = os.path.join(sr2_redis_data, file_name)
            command.append(file_path)
        command = ' '.join(command)
        net.ssh_execute(client, command)

    def cluster_clean(self, master=True, slave=True):
        if master:
            logger.info('clean redis conf, node conf, db data of master')
            for host in self.master_host_list:
                logger.info(' - {}'.format(host))
                client = net.get_ssh(host)
                self.remove_generated_config(client, self.master_port_list)
                self._remove_data(client, self.master_port_list)
                self._remove_node_conf(client, self.master_port_list)
                client.close()
        if slave and self.slave_host_list:
            logger.info('clean redis conf, node conf, db data of slave')
            for host in self.slave_host_list:
                logger.info(' - {}'.format(host))
                client = net.get_ssh(host)
                self.remove_generated_config(client, self.slave_port_list)
                self._remove_data(client, self.slave_port_list)
                self._remove_node_conf(client, self.slave_port_list)
                client.close()

    def update_ip_port(self):
        logger.debug('update ip port')
        self.cluster_id = config.get_cur_cluster_id()
        self.master_host_list = config.get_master_host_list(self.cluster_id)
        self.slave_host_list = config.get_slave_host_list(self.cluster_id)
        self.master_port_list = config.get_master_port_list(self.cluster_id)
        self.slave_port_list = config.get_slave_port_list(self.cluster_id)
        m_host_list = self.master_host_list
        s_host_list = self.slave_host_list
        self.all_host_list = list(set(m_host_list + s_host_list))

    def check_conf_file_exist(self, hosts, ports):
        path_of_fb = config.get_path_of_fb(self.cluster_id)
        sr2_redis_conf = path_of_fb['sr2_redis_conf']
        for host in hosts:
            conf_path_list = []
            res = True
            count = 0
            for port in ports:
                if count > 100:
                    client = net.get_ssh(host)
                    res = res and net.is_exist_files(client, conf_path_list)
                    client.close()
                    count = 0
                    conf_path_list = []
                count += 1
                conf_path = '{}/redis-{}.conf'.format(sr2_redis_conf, port)
                conf_path_list.append(conf_path)
            client = net.get_ssh(host)
            res = res and net.is_exist_files(client, conf_path_list)
            client.close()
            if not res:
                msg = "conf file not exist at '{}'".format(host)
                raise ClusterRedisError(msg)

    def start_redis_process(self, profile=False, master=True, slave=True):
        """ Start redis process
        """
        logger.debug('start_redis_process.')
        if master:
            self.check_conf_file_exist(
                self.master_host_list,
                self.master_port_list
            )
        if slave:
            self.check_conf_file_exist(
                self.slave_host_list,
                self.slave_port_list
            )
        current_time = time.strftime("%Y%m%d-%H%M", time.gmtime())
        if master:
            m_port = self.master_port_list
            for host in self.master_host_list:
                logger.info('Starting master nodes : {} : {} ...'.format(
                    host,
                    '|'.join(list(map(str, m_port)))
                ))
                self.run_redis_process(host, m_port, profile, current_time)
        if slave:
            s_port = self.slave_port_list
            for host in self.slave_host_list:
                logger.info('Starting slave nodes : {} : {} ...'.format(
                    host,
                    '|'.join(list(map(str, s_port)))
                ))
                self.run_redis_process(host, s_port, profile, current_time)

    def run_redis_process(self, host, ports, profile, current_time):
        logger.debug('run_redis_process')
        path_of_fb = config.get_path_of_fb(self.cluster_id)
        sr2_redis_bin = path_of_fb['sr2_redis_bin']
        sr2_redis_conf = path_of_fb['sr2_redis_conf']
        sr2_redis_log = path_of_fb['sr2_redis_log']
        sr2_redis_lib = path_of_fb['sr2_redis_lib']
        lib_path = config.get_ld_library_path(self.cluster_id)

        # create log directory
        client = net.get_ssh(host)
        command = 'mkdir -p {}'.format(sr2_redis_log)
        net.ssh_execute(client, command)

        # make env
        env_cmd = [
            'GLOBIGNORE=*;',
            'export LD_LIBRARY_PATH={};'.format(lib_path['ld_library_path']),
            'export DYLD_LIBRARY_PATH={};'.format(
                lib_path['dyld_library_path']
            ),
            'export SR2_REDIS_BIN={};'.format(sr2_redis_bin),
            'export SR2_REDIS_CONF={};'.format(sr2_redis_conf),
            'export SR2_REDIS_LOG={};'.format(sr2_redis_log),
            'export SR2_REDIS_LIB={};'.format(sr2_redis_lib),
        ]
        if profile:
            env_cmd.append(
                'MALLOC_CONF=prof_leak:true,lg_prof_sample:0,prof_final:true;'
            )
            env_cmd.append(
                'LD_PRELOAD=$SR2_REDIS_LIB/native/libjemalloc.so;'
            )

        run_cmd = '$SR2_REDIS_BIN/redis-server'
        command = [' '.join(env_cmd)]
        count = 0
        for port in ports:
            if count > 100:
                command = ' '.join(command)
                client = net.get_ssh(host)
                net.ssh_execute(client, command)
                client.close()
                command = [' '.join(env_cmd)]
                count = 0
            count += 1
            conf_file_name = 'redis-{}.conf'.format(port)
            log_file_name = 'servers-{}-{}.log'.format(current_time, port)
            command.append('({} {} >> {} 2>&1) &'.format(
                run_cmd,
                '$SR2_REDIS_CONF/{}'.format(conf_file_name),
                '$SR2_REDIS_LOG/{}'.format(log_file_name),
            ))
        command = ' '.join(command)
        client = net.get_ssh(host)
        net.ssh_execute(client, command)
        client.close()

    def ensure_cluster_exist(self):
        logger.debug('ensure_cluster_exist')
        logger.info('Check cluster exist...')
        hosts = self.all_host_list
        for host in hosts:
            logger.info(' - {}'.format(host))
            deploy_state = DeployUtil().get_state(self.cluster_id, host)
            if deploy_state != DEPLOYED:
                raise ClusterNotExistError(self.cluster_id, host=host)
        logger.info('OK')

    @staticmethod
    def _get_ip_port_dict_using_cluster_nodes_cmd():
        def mute_formatter(outs):
            pass

        outs = RedisCliUtil.command(
            sub_cmd='cluster nodes',
            formatter=mute_formatter)
        lines = outs.splitlines()
        d = {}
        for line in lines:
            rows = line.split(' ')
            addr = rows[1]
            if 'connected' in rows:
                (host, port) = addr.split(':')
                if host not in d:
                    d[host] = [port]
                else:
                    d[host].append(port)
        return d

    def _get_master_slave_pair_list(self):
        m_count = len(self.master_host_list)
        m_port_count = len(self.master_port_list)
        ret = []
        for i, m_host in enumerate(self.master_host_list):
            for j, s_port in enumerate(self.slave_port_list):
                master_port_idx = j % m_port_count
                slave_host_idx = (i + 1 + j // m_port_count) % m_count
                ret.append([
                    m_host,  # master host
                    self.master_port_list[master_port_idx],  # master port
                    self.slave_host_list[slave_host_idx],  # slave host
                    s_port  # slave port
                ])
        return ret

    @staticmethod
    def _replicate_thread(m_ip, m_port, s_ip, s_port, fail_list):
        logger.info('replicate [M] %s %s - [S] %s %s' % (
            m_ip, m_port, s_ip, s_port))
        try:
            m_ip = net.get_ip(m_ip)
            s_ip = net.get_ip(s_ip)
            trib.replicate(m_ip, m_port, s_ip, s_port)
        except Exception as ex:
            m_addr = '{}:{}'.format(m_ip, m_port)
            s_addr = '{}:{}'.format(s_ip, s_port)
            msg = 'Fail replicate [M] {} - [S] {}'.format(m_addr, s_addr)
            logger.error('\n'.join([msg, str(ex)]))
            fail_list.append((m_ip, m_port, s_ip, s_port))

    def replicate(self):
        threads = []
        fail_list = []
        pair_list = self._get_master_slave_pair_list()
        for m_ip, m_port, s_ip, s_port in pair_list:
            t = Thread(
                target=Center._replicate_thread,
                args=(m_ip, m_port, s_ip, s_port, fail_list)
            )
            threads.append(t)
        for x in threads:
            time.sleep(0.02)
            x.start()
        for x in threads:
            x.join()
        logger.info('{} / {} replicate completion'.format(
            len(threads) - len(fail_list),
            len(threads)
        ))

    def cli_config_get(self, key, host, port):
        lib_path = config.get_ld_library_path(self.cluster_id)
        path_of_fb = config.get_path_of_fb(self.cluster_id)
        sr2_redis_bin = path_of_fb['sr2_redis_bin']
        sr2_redis_conf = path_of_fb['sr2_redis_conf']
        sr2_redis_log = path_of_fb['sr2_redis_log']
        sr2_redis_lib = path_of_fb['sr2_redis_lib']
        env_cmd = [
            'GLOBIGNORE=*;',
            'export LD_LIBRARY_PATH={};'.format(lib_path['ld_library_path']),
            'export DYLD_LIBRARY_PATH={};'.format(
                lib_path['dyld_library_path']
            ),
            'export SR2_REDIS_BIN={};'.format(sr2_redis_bin),
            'export SR2_REDIS_CONF={};'.format(sr2_redis_conf),
            'export SR2_REDIS_LOG={};'.format(sr2_redis_log),
            'export SR2_REDIS_LIB={};'.format(sr2_redis_lib),
        ]
        redis_cli_cmd = os.path.join(sr2_redis_bin, 'redis-cli')
        sub_cmd = 'config get "{}"'.format(key)
        command = '{} {} -h {} -p {} {}'.format(
            ' '.join(env_cmd),
            redis_cli_cmd,
            host,
            port,
            sub_cmd,
        )
        try:
            stdout = subprocess.check_output(command, shell=True)
            stdout = stdout.decode('utf-8')
            return stdout.strip().split()[1]
        except IndexError as ex:
            logger.debug("Cannot get value by key '{}'".format(key))
            return False
        except subprocess.CalledProcessError as ex:
            logger.debug(ex)
            return False

    def cli_config_set_all(self, key, value, hosts, ports):
        lib_path = config.get_ld_library_path(self.cluster_id)
        path_of_fb = config.get_path_of_fb(self.cluster_id)
        sr2_redis_bin = path_of_fb['sr2_redis_bin']
        sr2_redis_conf = path_of_fb['sr2_redis_conf']
        sr2_redis_log = path_of_fb['sr2_redis_log']
        sr2_redis_lib = path_of_fb['sr2_redis_lib']
        env_cmd = [
            'GLOBIGNORE=*;',
            'export LD_LIBRARY_PATH={};'.format(lib_path['ld_library_path']),
            'export DYLD_LIBRARY_PATH={};'.format(
                lib_path['dyld_library_path']
            ),
            'export SR2_REDIS_BIN={};'.format(sr2_redis_bin),
            'export SR2_REDIS_CONF={};'.format(sr2_redis_conf),
            'export SR2_REDIS_LOG={};'.format(sr2_redis_log),
            'export SR2_REDIS_LIB={};'.format(sr2_redis_lib),
        ]
        redis_cli_cmd = os.path.join(sr2_redis_bin, 'redis-cli')
        sub_cmd = 'config set {} {}'.format(key, value)
        success = True
        for host in hosts:
            for port in ports:
                command = '{} {} -h {} -p {} {}'.format(
                    ' '.join(env_cmd),
                    redis_cli_cmd,
                    host,
                    port,
                    sub_cmd,
                )
                stdout = subprocess.check_output(command, shell=True)
                if stdout.decode('utf-8').strip() != 'OK':
                    success = False
        return success

    def get_cluster_nodes(self):
        lib_path = config.get_ld_library_path(self.cluster_id)
        path_of_fb = config.get_path_of_fb(self.cluster_id)
        sr2_redis_bin = path_of_fb['sr2_redis_bin']
        env_cmd = [
            'GLOBIGNORE=*;',
            'export LD_LIBRARY_PATH={};'.format(lib_path['ld_library_path']),
            'export DYLD_LIBRARY_PATH={};'.format(
                lib_path['dyld_library_path']
            ),
        ]
        redis_cli_cmd = os.path.join(sr2_redis_bin, 'redis-cli')
        sub_cmd = 'cluster nodes 2>&1'
        t = 2
        for host in self.master_host_list:
            for port in self.master_port_list:
                command = '{} timeout {} {} -h {} -p {} {}'.format(
                    ' '.join(env_cmd),
                    t,
                    redis_cli_cmd,
                    host,
                    port,
                    sub_cmd,
                )
                try:
                    ret = subprocess.check_output(command, shell=True)
                    return utils.to_str(ret)
                except Exception as ex:
                    logger.debug(ex)
        for host in self.slave_host_list:
            for port in self.slave_port_list:
                command = '{} timeout {} {} -h {} -p {} {}'.format(
                    ' '.join(env_cmd),
                    t,
                    redis_cli_cmd,
                    host,
                    port,
                    sub_cmd,
                )
                try:
                    ret = subprocess.check_output(command, shell=True)
                    return utils.to_str(ret)
                except Exception as ex:
                    logger.debug(ex)
        msg = 'All redis is disconnected or paused.'
        raise ClusterRedisError(msg)

    def ping(self, addr, t=0.5, c=3):
        """ping to redis
        return exit status
        0: PONG
        1: connection refused
        124: timeout
        """
        host, port = addr.split(':')
        lib_path = config.get_ld_library_path(self.cluster_id)
        path_of_fb = config.get_path_of_fb(self.cluster_id)
        sr2_redis_bin = path_of_fb['sr2_redis_bin']
        env_cmd = [
            'GLOBIGNORE=*;',
            'export LD_LIBRARY_PATH={};'.format(lib_path['ld_library_path']),
            'export DYLD_LIBRARY_PATH={};'.format(
                lib_path['dyld_library_path']
            ),
        ]
        redis_cli_cmd = os.path.join(sr2_redis_bin, 'redis-cli')
        sub_cmd = 'ping > /dev/null 2>&1'
        exit_code = -1
        while c > 0 and exit_code is not 0:
            c -= 1
            command = '{} timeout {} {} -h {} -p {} {}'.format(
                ' '.join(env_cmd),
                t,
                redis_cli_cmd,
                host,
                port,
                sub_cmd,
            )
            exit_code = subprocess.call(command, shell=True)
            logger.debug('ping {}: {}'.format(addr, exit_code))
        return exit_code

    def run_failover(self, addr, take_over=False):
        logger.debug('run failover {}'.format(addr))
        host, port = addr.split(':')
        lib_path = config.get_ld_library_path(self.cluster_id)
        path_of_fb = config.get_path_of_fb(self.cluster_id)
        sr2_redis_bin = path_of_fb['sr2_redis_bin']
        env_cmd = [
            'GLOBIGNORE=*;',
            'export LD_LIBRARY_PATH={};'.format(lib_path['ld_library_path']),
            'export DYLD_LIBRARY_PATH={};'.format(
                lib_path['dyld_library_path']
            ),
        ]
        redis_cli_cmd = os.path.join(sr2_redis_bin, 'redis-cli')
        sub_cmd = 'cluster failover'
        if take_over:
            sub_cmd += ' takeover'
        command = '{} {} -h {} -p {} {}'.format(
            ' '.join(env_cmd),
            redis_cli_cmd,
            host,
            port,
            sub_cmd,
        )
        stdout = subprocess.check_output(command, shell=True)
        stdout = utils.to_str(stdout)
        logger.debug('{} failover stdout: {}'.format(addr, stdout))
        return stdout.strip()

    def get_master_obj_list(self):
        """get object list of master hosts
        return [
            {
                node_id: string
                addr: string(host:port)
                status: string(connected / disconnected / paused
                slaves: [
                    {
                        node_id: string
                        addr: string(host:port)
                        status: string(connected / disconnected / paused
                    }
                ]
            }
        ]
        """
        cluster_nodes = self.get_cluster_nodes()
        logger.debug('result of cluster nodes: {}'.format(cluster_nodes))
        nodes_info = cluster_nodes.split('\n')
        master_nodes_info = []
        slave_nodes_info = []
        for line in nodes_info:
            if 'master' in line:
                master_nodes_info.append(line)
            if 'slave' in line:
                slave_nodes_info.append(line)

        if len(master_nodes_info) <= 1:
            raise ClusterRedisError("Need to create cluster")

        master_node_list = []
        for line in master_nodes_info:
            status = 'disconnected' if 'disconnected' in line else 'connected'
            splited = line.split()
            if status == 'connected':
                exit_code = self.ping(splited[1])
                if exit_code == 124:
                    status = 'paused'
            master_node_list.append({
                "node_id": splited[0],
                "addr": splited[1],
                "status": status,
                "slaves": []
            })
        master_node_list.sort(key=lambda node: node['addr'])

        for line in slave_nodes_info:
            status = 'disconnected' if 'disconnected' in line else 'connected'
            splited = line.split()
            if status == 'connected':
                exit_code = self.ping(splited[1])
                if exit_code == 124:
                    status = 'paused'
            for master_node in master_node_list:
                if master_node["node_id"] in line:
                    master_node["slaves"].append({
                        "node_id": splited[0],
                        "addr": splited[1],
                        "status": status
                    })
        master_node_list.sort(key=lambda node: node['addr'].split(':')[1])
        for master in master_node_list:
            master["slaves"].sort(key=lambda node: node['addr'].split(':')[1])
        return master_node_list

    def check_all_master_have_alive_slave(self):
        master_obj_list = self.get_master_obj_list()
        slaves_for_failover = []
        for master in master_obj_list:
            no_slave = True
            for slave in master['slaves']:
                if slave['status'] == 'connected':
                    slaves_for_failover.append(slave['addr'])
                    no_slave = False
                    break
            if no_slave:
                msg = 'Not exist alive slave: {}'.format(master['addr'])
                raise ClusterRedisError(msg)
        return slaves_for_failover



import os
from setuptools import setup, find_packages


here = os.path.abspath(os.path.dirname(__file__))

requirements = []
with open('requirements.txt') as file_requirements:
    requirements = file_requirements.read().splitlines()

about = {}
with open(os.path.join(here, 'ltcli', '__version__.py'), 'r') as f:
    exec(f.read(), about)

setup(
    name=about['__title__'],
    version=about['__version__'],
    description=about['__description__'],
    author=about['__author__'],
    author_email=about['__email__'],
    url=about['__url__'],
    install_requires=requirements,
    packages=find_packages(exclude=['tests', 'docs', 'sql']),
    python_requires='>=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*',
    package_data={},
    zip_safe=False,
    entry_points={
        'console_scripts': [
            'ltcli = ltcli.cli_main:main'
        ]
    },
    classifiers=[
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ]
)



import os

from ltcli.log import logger
from ltcli import config
from ltcli import utils


def validate_id(cluster_id):
    if not isinstance(cluster_id, int):
        return False
    if cluster_id < 1:
        return False
    return True


def get_cluster_list():
    base_directory = config.get_base_directory()
    buf = os.listdir(base_directory)
    buf = filter(lambda x: x.startswith('cluster_'), buf)
    buf = filter(lambda x: utils.is_number(str(x[8:])), buf)
    buf = list(map(lambda x: int(x[8:]), buf))
    cluster_list = []
    for cluster_id in buf:
        cluster_dir = 'cluster_{}'.format(cluster_id)
        cluster_path = os.path.join(base_directory, cluster_dir)
        if not os.path.isfile(os.path.join(cluster_path, '.deploy.state')):
            cluster_list.append(int(cluster_id))
    cluster_list.sort()
    return cluster_list


def convert_list_2_seq(ports):
    logger.debug('ports: {}'.format(ports))
    ports.sort()
    logger.debug('sorted ports: {}'.format(ports))
    ret = []
    s = ports[0]
    pre = ports[0] - 1
    for port in ports:
        if pre != port - 1:
            if s != pre:
                ret.append('$(seq {} {})'.format(s, pre))
            else:
                ret.append(s)
            s = port
        pre = port
    if s != pre:
        ret.append('$(seq {} {})'.format(s, pre))
    else:
        ret.append(s)
    logger.debug('converted: {}'.format(ret))
    return ret



DEFAULT = '\033[39m'
BLACK = '\033[30m'
RED = '\033[31m'
GREEN = '\033[32m'
YELLOW = '\033[33m'
BLUE = '\033[34m'
MAGENTA = '\033[35m'
CYAN = '\033[36m'
LIGHTGRAY = '\033[37m'
DARKGRAY = '\033[90m'
LIGHTRED = '\033[91m'
LIGHTGREEN = '\033[92m'
LIGHTYELLOW = '\033[93m'
LIGHTBLUE = '\033[94m'
LIGHTMAGENTA = '\033[95m'
LIGHTCYAN = '\033[96m'
WHITE = '\033[97m'

ENDC = '\033[0m'
BOLD = '\033[1m'
UNDERLINE = '\033[4m'


def green(msg):
    return GREEN + msg + ENDC


def red(msg):
    return RED + msg + ENDC


def yellow(msg):
    return YELLOW + msg + ENDC


def blue(msg):
    return BLUE + msg + ENDC


def cyan(msg):
    return CYAN + msg + ENDC


def magenta(msg):
    return MAGENTA + msg + ENDC


def white(msg):
    return WHITE + msg + ENDC



import os
import socket
import time
import re
import subprocess
import shutil

import yaml

from ltcli.log import logger
from ltcli.exceptions import (
    YamlSyntaxError,
    PropsSyntaxError,
    ClusterIdError,
    ClusterNotExistError,
    PropsError,
    PropsKeyError,
)


def get_local_ip():
    return socket.gethostbyname(socket.gethostname())


def get_local_ip_list():
    return [
        socket.gethostbyname(socket.gethostname()),
        socket.gethostname(),
        'localhost',
        '127.0.0.1'
    ]


def get_root_of_cli_config():
    p = os.environ['FBPATH']
    if not os.path.exists(p):
        os.path.mkdir(p)
    return os.environ['FBPATH']


def get_cur_cluster_id(allow_empty_id=False):
    """Get cur cluster id

    :return: cluster #
    """
    root_of_cli_config = get_root_of_cli_config()
    head_path = os.path.join(root_of_cli_config, 'HEAD')
    if not os.path.exists(head_path):
        with open(head_path, 'w') as fd:
            fd.writelines(str(-1))
    with open(head_path, 'r') as fd:
        line = fd.readline().strip()
        cluster_id = line
    if line == '-1' and allow_empty_id:
        return -1
    if not is_number(line):
        raise ClusterIdError(cluster_id)
    cluster_id = int(cluster_id)
    base_directory = get_base_directory()
    buf = os.listdir(base_directory)
    buf = filter(lambda x: x.startswith('cluster_'), buf)
    buf = filter(lambda x: is_number(str(x[8:])), buf)
    buf = list(map(lambda x: int(x[8:]), buf))
    cluster_list = []
    for cid in buf:
        cluster_dir = 'cluster_{}'.format(cid)
        cluster_path = os.path.join(base_directory, cluster_dir)
        if not os.path.isfile(os.path.join(cluster_path, '.deploy.state')):
            cluster_list.append(int(cid))
    if cluster_id not in cluster_list:
        raise ClusterNotExistError(cluster_id)
    return cluster_id


def get_repo_cluster_path(cluster_id=-1):
    """Get repo cluster path

    Let cur cluster id is 1, then return ${root_of_cli_config}/clusters/1

    :return: repo cluster path
    """
    root_of_cli_config = get_root_of_cli_config()
    if cluster_id < 0:
        cluster_id = get_cur_cluster_id()
    return os.path.join(root_of_cli_config, 'clusters', str(cluster_id))


def get_repo_cluster_template_path():
    """Get repo cluster template path

    Let cur cluster id is 1,
    then return ${root_of_cli_config}/clusters/template

    :return: repo cluster template path
    """
    root_of_cli_config = get_root_of_cli_config()
    return os.path.join(root_of_cli_config, 'clusters', 'template')


def get_config(cluster_id=-1, template=False):
    """Get config

    :param cluster_id: target cluster #
    :param template: If true, get config from template
    :return: config dict
    """
    if template:
        cur_cluster_path = get_repo_cluster_template_path()
    else:
        cur_cluster_path = get_repo_cluster_path(cluster_id)
    yml_path = os.path.join(cur_cluster_path, 'config.yaml')
    with open(yml_path, 'r') as f:
        stream = ''.join(f.readlines())
        fb_config = yaml.load(stream, Loader=yaml.FullLoader)
        return fb_config
    assert False
    return None


def get_node_ip_list(cluster_id=-1):
    """Get node ip list from config

    :param cluster_id: target cluster #
    :return: list of ip
    """
    if cluster_id == -1:
        cluster_id = get_cur_cluster_id()
    path_of_fb = get_path_of_fb(cluster_id)
    props_path = path_of_fb['redis_properties']
    key = 'sr2_redis_master_hosts'
    nodes = get_props(props_path, key, [])
    ip_list = []
    for node in nodes:
        ip = socket.gethostbyname(node)
        ip_list.append(ip)
    return ip_list


def get_master_host_list(cluster_id=None):
    if cluster_id is None:
        cluster_id = get_cur_cluster_id()
    path_of_fb = get_path_of_fb(cluster_id)
    props_path = path_of_fb['redis_properties']
    key = 'sr2_redis_master_hosts'
    hosts = get_props(props_path, key, [])
    return hosts


def get_master_ip_list(cluster_id=None):
    if cluster_id is None:
        cluster_id = get_cur_cluster_id()
    hosts = get_master_host_list(cluster_id)
    ip_list = []
    for host in hosts:
        ip = socket.gethostbyname(host)
        ip_list.append(ip)
    return ip_list


def get_slave_host_list(cluster_id=None):
    if cluster_id is None:
        cluster_id = get_cur_cluster_id()
    path_of_fb = get_path_of_fb(cluster_id)
    props_path = path_of_fb['redis_properties']
    key = 'sr2_redis_slave_hosts'
    hosts = get_props(props_path, key, [])
    return hosts


def get_slave_ip_list(cluster_id=None):
    if cluster_id is None:
        cluster_id = get_cur_cluster_id()
    hosts = get_slave_host_list(cluster_id)
    ip_list = []
    for host in hosts:
        ip = socket.gethostbyname(host)
        ip_list.append(ip)
    return ip_list


def get_replicas(cluster_id=None):
    """Get replicas using port count for master and slave

    :return: replicas
    """
    logger.debug('get_replicas')
    if cluster_id is None:
        cluster_id = get_cur_cluster_id()
    m_len = len(get_master_port_list(cluster_id))
    if m_len <= 0:
        raise PropsKeyError('sr2_redis_master_hosts')
    s_len = len(get_slave_port_list(cluster_id))
    if s_len == 0:
        ret = 0
        logger.debug('return replicas: {}'.format(ret))
        return ret
    if s_len % m_len is not 0:
        msg = [
            'The number of slaves should be multiple values ',
            'with the number of masters.\n',
            'master: {}\n'.format(m_len),
            'slave: {}'.format(s_len),
        ]
        raise PropsError(''.join(msg))
    ret = s_len // m_len
    logger.debug('return replicas: {}'.format(ret))
    return ret


def get_master_port_list(cluster_id=-1):
    """Get master port list

    :param cluster_id: target cluster #
    :return: master port list
    """
    if cluster_id == -1:
        cluster_id = get_cur_cluster_id()
    path_of_fb = get_path_of_fb(cluster_id)
    props_path = path_of_fb['redis_properties']
    key = 'sr2_redis_master_ports'
    ports = get_props(props_path, key, [])
    return ports


def get_slave_port_list(cluster_id=-1):
    """Get slave port list

    :return: slave port list
    """
    if cluster_id == -1:
        cluster_id = get_cur_cluster_id()
    path_of_fb = get_path_of_fb(cluster_id)
    props_path = path_of_fb['redis_properties']
    key = 'sr2_redis_slave_ports'
    ports = get_props(props_path, key, [])
    return ports


def is_slave_enabled():
    """Return slave enable or not using config

    :return: True | False
    """
    cluster_id = get_cur_cluster_id()
    path_of_fb = get_path_of_fb(cluster_id)
    props_path = path_of_fb['redis_properties']
    key = 'sr2_redis_slave_hosts'
    return is_key_enable(props_path, key)


def get_base_directory(expanduser=True):
    cli_config = get_cli_config()
    base_directory = cli_config['base_directory']
    if expanduser:
        base_directory = os.path.expanduser(base_directory)
    return base_directory


def get_tsr2_home(cluster_id=None):
    """Get tsr2 home path

    This is for deploy, copy redis.conf, backup remote logs, etc.

    :param cluster_id: target cluster #
    :return: tsr2 home path
    """
    if cluster_id is None:
        cluster_id = get_cur_cluster_id()
    base_directory = get_base_directory()
    tsr2_home = os.path.join(
        base_directory,
        'cluster_{}'.format(cluster_id),
        'tsr2-assembly-1.0.0-SNAPSHOT'
    )
    return tsr2_home


def get_sata_ssd_no(port, count, digit=2):
    if count < 1:
        return ''
    rest_number = str(port % count + 1).zfill(digit)
    return rest_number


def get_ssd_disk_position(port, digit=2):
    user = os.environ['USER']
    cluster_id = get_cur_cluster_id()
    path_of_fb = get_path_of_fb(cluster_id)
    props_path = path_of_fb['redis_properties']
    ssd_count = int(get_props(props_path, 'ssd_count', 0))
    redis_data_prefix = get_props(props_path, 'sr2_redis_data')
    flash_db_path_prefix = get_props(props_path, 'sr2_flash_db_path')
    redis_data = os.path.join(
        redis_data_prefix + get_sata_ssd_no(port, ssd_count, digit),
        user,
    )
    flash_db_path = os.path.join(
        flash_db_path_prefix + get_sata_ssd_no(port, ssd_count, digit),
        user,
        'db/db-{port}'.format(port=port),
    )
    redis_dump = os.path.join(redis_data, 'dump')
    return {
        'sr2_redis_data': redis_data,
        'sr2_flash_db_path': flash_db_path,
        'sr2_redis_dump': redis_dump,
        'sr2_redis_db_path': flash_db_path,
    }


def get_path_of_fb(cluster_id):
    base_directory = get_base_directory()
    cluster_dir = 'cluster_{}'.format(cluster_id)
    cluster_path = os.path.join(base_directory, cluster_dir)
    cluster_backup_path = os.path.join(base_directory, 'backup')
    release_path = os.path.join(base_directory, 'releases')
    sr2_home = get_tsr2_home(cluster_id)
    conf_path = os.path.join(sr2_home, 'conf')
    redis_properties = os.path.join(conf_path, 'redis.properties')
    master_template = os.path.join(conf_path, 'redis-master.conf.template')
    slave_template = os.path.join(conf_path, 'redis-slave.conf.template')
    thrift_properties = os.path.join(conf_path, 'thriftserver.properties')
    sr2_redis_log = os.path.join(sr2_home, 'logs', 'redis')
    sr2_redis_conf = os.path.join(sr2_home, 'conf', 'redis')
    sr2_redis_conf_temp = os.path.join(sr2_home, 'conf', 'temp')
    sr2_redis_bin = os.path.join(sr2_home, 'bin')
    sr2_redis_lib = os.path.join(sr2_home, 'lib')

    return {
        'base_directory': base_directory,
        'cluster_path': cluster_path,
        'cluster_backup_path': cluster_backup_path,
        'conf_path': conf_path,
        'release_path': release_path,
        'redis_properties': redis_properties,
        'master_template': master_template,
        'slave_template': slave_template,
        'thrift_properties': thrift_properties,
        'sr2_redis_home': sr2_home,
        'sr2_redis_log': sr2_redis_log,
        'sr2_redis_conf': sr2_redis_conf,
        'sr2_redis_conf_temp': sr2_redis_conf_temp,
        'sr2_redis_bin': sr2_redis_bin,
        'sr2_redis_lib': sr2_redis_lib,
    }


def get_env_dict(ip, port):
    """Collection of env

    Build env from config and return it as dict type.

    :param ip: ip
    :param port: port
    :return: dict
    """

    ssd_disk_position = get_ssd_disk_position(port)
    cluster_id = get_cur_cluster_id()
    path_of_fb = get_path_of_fb(cluster_id)
    return {
        'sr2_redis_home': path_of_fb['sr2_redis_home'],
        'sr2_redis_bin': path_of_fb['sr2_redis_bin'],
        'sr2_redis_lib': path_of_fb['sr2_redis_lib'],
        'sr2_redis_conf': path_of_fb['sr2_redis_conf'],
        'sr2_redis_log': path_of_fb['sr2_redis_log'],
        'sr2_redis_data': ssd_disk_position['sr2_redis_data'],
        'sr2_redis_dump': ssd_disk_position['sr2_redis_dump'],
        'sr2_flash_db_path': ssd_disk_position['sr2_flash_db_path'],
        'sr2_redis_db_path': ssd_disk_position['sr2_redis_db_path'],
        'ld_library_path': ':'.join([
            '{}/native'.format(path_of_fb['sr2_redis_lib']),
            '/usr/lib64',
        ]),
        'dyld_library_path': ':'.join([
            '{}/native'.format(path_of_fb['sr2_redis_lib']),
            '/usr/lib64',
        ]),
        'sr2_redis_host': ip,
        'sr2_redis_port': port,
    }


def get_ld_library_path(cluster_id):
    path_of_fb = get_path_of_fb(cluster_id)
    return {
        'ld_library_path': ':'.join([
            '$LD_LIBRARY_PATH',
            '{}/native'.format(path_of_fb['sr2_redis_lib']),
            '/usr/lib64',
        ]),
        'dyld_library_path': ':'.join([
            '$DYLD_LIBRARY_PATH',
            '{}/native'.format(path_of_fb['sr2_redis_lib']),
            '/usr/lib64',
        ]),
    }


def get_path_of_cli(cluster_id):
    root_of_cli_config = os.path.expanduser(get_root_of_cli_config())
    conf_backup_path = os.path.join(root_of_cli_config, 'conf_backup')
    release_path = os.path.join(root_of_cli_config, 'releases')
    cluster_path = os.path.join(root_of_cli_config, 'clusters', str(cluster_id))
    conf_path = os.path.join(cluster_path, 'conf')
    redis_properteis = os.path.join(conf_path, 'redis.properties')

    return {
        'cli_config_root': root_of_cli_config,
        'conf_backup_path': conf_backup_path,
        'release_path': release_path,
        'cluster_path': cluster_path,
        'conf_path': conf_path,
        'redis_properties': redis_properteis,
    }


def reset_conf_of_cli(cluster_id, backup=False):
    logger.debug('reset conf of cli')
    path_of_fb = get_path_of_fb(cluster_id)
    path_of_cli = get_path_of_cli(cluster_id)
    if backup and os.path.isdir(path_of_cli['conf_path']):
        current_time = time.strftime("%Y%m%d%H%M%s", time.gmtime())
        conf_backup_dir = 'cluster_{}_conf_bak_{}'.format(
            cluster_id, current_time)
        conf_backup_path = path_of_cli['conf_backup_path']
        shutil.copytree(
            path_of_fb['conf_path'],
            os.path.join(
                conf_backup_path,
                conf_backup_dir))
        logger.debug("conf backup: '{}'".format(conf_backup_dir))
    if not os.path.isdir(path_of_cli['cluster_path']):
        logger.debug(
            "FileNotExisted: '{}'".format(
                path_of_cli['cluster_path']))
        os.mkdir(path_of_cli['cluster_path'])
        logger.debug("CreateDir: '{}'".format(path_of_cli['cluster_path']))
    if os.path.isdir(path_of_cli['conf_path']):
        shutil.rmtree(path_of_cli['conf_path'], ignore_errors=True)
        logger.debug("RemoveDir: '{}'".format(path_of_cli['conf_path']))
    shutil.copytree(path_of_fb['conf_path'], path_of_cli['conf_path'])
    msg = [
        'copy tree',
        "from '{}'".format(path_of_fb['conf_path']),
        "to '{}'".format(path_of_cli['conf_path']),
    ]
    logger.debug(' '.join(msg))


def get_cli_config():
    root_of_cli_config = get_root_of_cli_config()
    conf_path = os.path.join(root_of_cli_config, 'config')
    conf_exist = os.path.exists(conf_path)
    if not conf_exist:
        with open(conf_path, 'w') as f:
            f.writelines("base_directory:")
    with open(os.path.join(root_of_cli_config, 'config'), 'r') as f:
        stream = ''.join(f.readlines())
        cli_config = yaml.load(stream, Loader=yaml.FullLoader)
    return cli_config


def save_cli_config(cli_config):
    root_of_cli_config = get_root_of_cli_config()
    with open(os.path.join(root_of_cli_config, 'config'), 'w') as f:
        yaml.dump(cli_config, f, default_flow_style=False)


def is_key_enable(props_path, key):
    with open(props_path, 'r') as f:
        key = key.upper()
        lines = f.readlines()
        for i, line in enumerate(lines):
            p = re.compile(r'export {}=(\(.+\)|[^ \s\t\r\n\v\f]+)'.format(key))
            m = p.match(line)
            if m:
                return True
        return False


# FIXME: delete v1 flg after meeting
def make_key_enable(props_path, key, v1_flg=False):
    if is_key_enable(props_path, key):
        if not v1_flg:
            return
    with open(props_path, 'r') as f:
        buf = []
        key = key.upper()
        lines = f.readlines()
        for i, line in enumerate(lines):
            p = re.compile(r'export {}=(\(.+\)|[^ \s\t\r\n\v\f]+)'.format(key))
            m = p.search(line)
            if line.strip().startswith('#') and m:
                s = m.start()
                buf.append(line[s:])
                buf = buf + lines[i + 1:]
                break
            buf.append(line)
    with open(props_path, 'w') as f:
        f.write(''.join(buf))


def make_key_disable(props_path, key):
    with open(props_path, 'r') as f:
        buf = []
        key = key.upper()
        lines = f.readlines()
        for i, line in enumerate(lines):
            p = re.compile(r'export {}=(\(.+\)|[^ \s\t\r\n\v\f]+)'.format(key))
            m = p.match(line)
            if m:
                buf.append('#' + line)
                buf = buf + lines[i + 1:]
                break
            buf.append(line)
    with open(props_path, 'w') as f:
        f.write(''.join(buf))


def set_props(props_path, key, value):
    key = key.upper()
    if isinstance(value, type(str())):
        value = '"{}"'.format(value)
    # pylint: disable=unidiomatic-typecheck
    # for checking muitple type
    if type(value) in [type(list()), type(tuple()), type(set())]:
        def f(x):
            if isinstance(x, str) and not x.startswith('$'):
                return '"{}"'.format(x)
            return x
        value = list(map(f, value))
        value = list(map(str, value))
        value = "( {} )".format(' '.join(value))

    with open(props_path, 'r') as f:
        buf = []
        lines = f.readlines()
        for i, line in enumerate(lines):
            p = re.compile(r'export {}=(\(.+\)|[^ \s\t\r\n\v\f]+)'.format(key))
            m = p.match(line)
            if m:
                buf.append('export {}={}\n'.format(key, value))
                buf = buf + lines[i + 1:]
                break
            buf.append(line)
    with open(props_path, 'w') as f:
        f.write(''.join(buf))


def get_props(props_path, key, default=None):
    logger.debug('Get props key: {}, default: {}'.format(key, default))
    props = get_props_as_dict(props_path)
    logger.debug(props)
    try:
        return props[key]
    except KeyError:
        msg = [
            "Key error in props: '{}'".format(key),
            "get default '{}'".format(default),
        ]
        logger.debug(' '.join(msg))
        if default is None:
            raise PropsKeyError(key)
        return default
    except IOError:
        msg = [
            "Props file is not existed",
            "get default '{}'".format(default),
        ]
        if default is None:
            raise PropsError(' '.join(msg))
        return default


def get_props_as_dict(props_path):
    ret = dict()
    with open(props_path, 'r') as f:
        lines = f.readlines()
        for i, line in enumerate(lines):
            if line.strip().startswith('#'):
                continue
            p = re.compile(
                r'export [^ \s\t\r\n\v\f]+=(\(.+\)|[^ \s\t\r\n\v\f]+)')
            m = p.search(line)
            if not m:
                continue
            s = m.start()
            e = m.end()
            key, value = line[s:e + 1].replace('export ', '').split('=')
            value = value.strip()
            key = key.lower()
            p = re.compile(r'\(.*\)')
            m = p.match(value)
            try:
                if m:
                    cmd = [
                        'FBCLI_TMP_ENV={}'.format(value),
                        '&&',
                        'echo ${FBCLI_TMP_ENV[@]}'
                    ]
                    cmd = ' '.join(cmd)
                    logger.debug('subprocess cmd: {}'.format(cmd))
                    value = subprocess.check_output(cmd, shell=True)
                    value = to_str(value.strip())
                    logger.debug('subprocess result: {}'.format(value))
                    value = value.split(' ')
                    value = map(lambda x: int(x) if is_number(x) else x, value)
                    value = filter(lambda x: bool(x), value)
                    value = list(value)
                else:
                    cmd = 'echo {}'.format(value)
                    value = subprocess.check_output(cmd, shell=True)
                    value = to_str(value.strip())
                    value = int(value) if is_number(value) else value
                ret[key] = value
            except subprocess.CalledProcessError:
                raise PropsSyntaxError(value, i + 1)
    return ret


def get_deploy_history():
    file_path = os.path.join(get_root_of_cli_config(), 'deploy_history')
    default = {
        'hosts': ['127.0.0.1'],
        'master_count': 10,
        'replicas': 0,
        'ssd_count': 4,
        'prefix_of_db_path': '/nvme/data_',
    }
    if not os.path.exists(file_path):
        with open(file_path, 'w') as f:
            yaml.dump(default, f, default_flow_style=False)
    try:
        with open(file_path, 'r') as f:
            stream = ''.join(f.readlines())
            ret = yaml.load(stream, Loader=yaml.FullLoader)
        if ret is None:
            with open(file_path, 'w') as f:
                yaml.dump(default, f, default_flow_style=False)
        with open(file_path, 'r') as f:
            stream = ''.join(f.readlines())
            ret = yaml.load(stream, Loader=yaml.FullLoader)
        with open(file_path, 'w') as f:
            d_keys = default.keys()
            r_keys = ret.keys()
            for key in d_keys:
                if key not in r_keys:
                    ret[key] = default[key]
            yaml.dump(ret, f, default_flow_style=False)
        with open(file_path, 'r') as f:
            stream = ''.join(f.readlines())
            return yaml.load(stream, Loader=yaml.FullLoader)
    except yaml.scanner.ScannerError:
        raise YamlSyntaxError(file_path)


def save_deploy_history(history):
    file_path = os.path.join(get_root_of_cli_config(), 'deploy_history')
    with open(file_path, 'w') as f:
        yaml.dump(history, f, default_flow_style=False)


def ensure_master_host_not_changed(tmp_path):
    cluster_id = get_cur_cluster_id()
    path_of_fb = get_path_of_fb(cluster_id)
    props_path = path_of_fb['redis_properties']
    tmp_host_list = get_props(tmp_path, 'sr2_redis_master_hosts')
    master_host_list = get_props(props_path, 'sr2_redis_master_hosts')
    if sorted(master_host_list) != sorted(tmp_host_list):
        raise PropsError("Cannot edit 'SR2_REDIS_MASTER_HOSTS'")


def is_number(target):
    try:
        target = target.decode('utf-8')
    except AttributeError:
        pass
    return target.isdecimal()


def to_str(target):
    if isinstance(target, bytes):
        target = target.decode('utf-8')
    return str(target)


def get_tmp_thriftserver_props_path():
    fbpath = get_root_of_cli_config()
    return os.path.join(fbpath, '.tmp', 'thriftserver.properties')


def make_tmp_thriftserver_props():
    fbpath = get_root_of_cli_config()
    tmp_dir_path = os.path.join(fbpath, '.tmp')
    if not os.path.isdir(tmp_dir_path):
        os.mkdir(tmp_dir_path)
    tmp_ths_props_path = os.path.join(tmp_dir_path, 'thriftserver.properties')
    if not os.path.isfile(tmp_ths_props_path):
        with open(tmp_ths_props_path, 'w') as f:
            f.write('''#!/bin/bash
###############################################################################
# Common variables
SPARK_CONF=${SPARK_CONF:-$SPARK_HOME/conf}
SPARK_BIN=${SPARK_BIN:-$SPARK_HOME/bin}
SPARK_SBIN=${SPARK_SBIN:-$SPARK_HOME/sbin}
SPARK_LOG=${SPARK_LOG:-$SPARK_HOME/logs}

SPARK_METRICS=${SPARK_CONF}/metrics.properties
SPARK_UI_PORT=${SPARK_UI_PORT:-14050}
EXECUTERS=2
EXECUTER_CORES=6

#HIVE_HOST=${HIVE_HOST:-localhost}
#HIVE_PORT=${HIVE_PORT:-13000}

COMMON_CLASSPATH=$(find $SR2_LIB -name 'tsr2*' -o -name 'spark-r2*' -o -name '*jedis*' -o -name 'commons*' -o -name 'jdeferred*' | tr '\n' ':')

###############################################################################
# Driver
DRIVER_MEMORY=2g
DRIVER_CLASSPATH=$COMMON_CLASSPATH

###############################################################################
# Execute
EXECUTOR_MEMORY=2g
EXECUTOR_CLASSPATH=$COMMON_CLASSPATH

###############################################################################
# Thrift Server logs
EVENT_LOG_ENABLED=false
EVENT_LOG_DIR=/nvdrive0/thriftserver-event-logs
EVENT_LOG_ROLLING_DIR=/nvdrive0/thriftserver-event-logs-rolling
EVENT_LOG_SAVE_MIN=60
EXTRACTED_EVENT_LOG_SAVE_DAY=5
SPARK_LOG_SAVE_MIN=2000
##############''')



import os
from functools import reduce
import time

from ltcli import color
from ltcli import config
from ltcli import cluster_util
from ltcli.center import Center
from ltcli.log import logger
from ltcli.rediscli_util import RedisCliUtil
from ltcli.redistrib2.custom_trib import rebalance_cluster_cmd
from ltcli.exceptions import (
    ClusterIdError,
    ClusterNotExistError,
    LightningDBError,
    ClusterRedisError
)


def _change_cluster(cluster_id):
    if not isinstance(cluster_id, int):
        raise ClusterIdError(cluster_id)
    root_of_cli_config = config.get_root_of_cli_config()
    head_path = os.path.join(root_of_cli_config, 'HEAD')
    cluster_list = cluster_util.get_cluster_list()
    if cluster_id not in cluster_list + [-1]:
        raise ClusterNotExistError(cluster_id)
    with open(head_path, 'w') as fd:
        fd.write('%s' % cluster_id)


class Cluster(object):
    """This is cluster command
    """

    def __init__(self, print_mode='screen'):
        self._print_mode = print_mode

    def stop(self, force=False, master=True, slave=True):
        """Stop cluster
        """
        if not isinstance(force, bool):
            logger.error("option '--force' can use only 'True' or 'False'")
            return
        if not isinstance(master, bool):
            logger.error("option '--master' can use only 'True' or 'False'")
            return
        if not isinstance(slave, bool):
            logger.error("option '--slave' can use only 'True' or 'False'")
            return
        center = Center()
        center.update_ip_port()
        success = center.check_hosts_connection()
        if not success:
            return
        center.stop_redis(force, master=master, slave=slave)

    def start(self, profile=False, master=True, slave=True):
        """Start cluster
        """
        logger.debug("command 'cluster start'")
        if not isinstance(profile, bool):
            logger.error("option '--profile' can use only 'True' or 'False'")
            return
        if not isinstance(master, bool):
            logger.error("option '--master' can use only 'True' or 'False'")
            return
        if not isinstance(slave, bool):
            logger.error("option '--slave' can use only 'True' or 'False'")
            return
        center = Center()
        center.update_ip_port()
        success = center.check_hosts_connection()
        if not success:
            return
        center.ensure_cluster_exist()
        if master:
            master_alive_count = center.get_alive_master_redis_count()
            if master_alive_count > 0:
                msg = [
                    'Fail to start master nodes... ',
                    'Must be checked running master processes!\n',
                    'We estimate that ',
                    "redis 'MASTER' processes is {}".format(master_alive_count)
                ]
                raise LightningDBError(11, ''.join(msg))
        slave_alive_count = center.get_alive_slave_redis_count()
        if slave:
            if slave_alive_count > 0:
                msg = [
                    'Fail to start slave nodes... ',
                    'Must be checked running slave processes!\n',
                    'We estimate that ',
                    "redis 'SLAVE' processes is {}".format(slave_alive_count)
                ]
                raise LightningDBError(12, ''.join(msg))
        center.backup_server_logs(master=master, slave=slave)
        center.create_redis_data_directory()

        # equal to cluster.configure()
        center.configure_redis()
        center.sync_conf(show_result=True)

        center.start_redis_process(profile, master=master, slave=slave)
        center.wait_until_all_redis_process_up(master=master, slave=slave)

    def create(self, yes=False):
        """Create cluster

        Before create cluster, all redis should be started.
        """
        center = Center()
        center.update_ip_port()
        success = center.check_hosts_connection()
        if not success:
            return

        m_count = len(center.master_host_list) * len(center.master_port_list)
        if m_count < 3:
            msg = [
                'To create cluster, ',
                '3 master processes should be included at least.',
            ]
            raise ClusterRedisError(''.join(msg))

        # if need to cluster start
        alive_count = center.get_alive_all_redis_count()
        my_alive_count = center.get_alive_all_redis_count(check_owner=True)
        if alive_count != my_alive_count:
            raise ClusterRedisError('The port range is already taken.')
        all_count = len(center.all_host_list)
        if alive_count < all_count:
            logger.debug('cluster start in create')
            # init
            center.backup_server_logs()
            center.create_redis_data_directory()

            # cluster configure
            center.configure_redis()
            center.sync_conf(show_result=True)

            # cluster start
            center.start_redis_process()
            center.wait_until_all_redis_process_up()

        key = 'cluster-node-timeout'
        m_hosts = center.master_host_list
        m_ports = center.master_port_list
        origin_m_value = center.cli_config_get(key, m_hosts[0], m_ports[0])
        if not origin_m_value:
            msg = "RedisConfigKeyError(master): '{}'".format(key)
            logger.warning(msg)
        s_hosts = center.slave_host_list
        s_ports = center.slave_port_list
        if s_hosts and s_ports:
            origin_s_value = center.cli_config_get(key, s_hosts[0], s_ports[0])
            if not origin_s_value:
                msg = "RedisConfigKeyError(slave): '{}'".format(key)
                logger.warning(msg)
        if origin_m_value:
            # cli config set cluster-node-timeout 2000
            logger.debug('set cluster node time out 2000 for create')
            center.cli_config_set_all(key, '2000', m_hosts, m_ports)
            if s_hosts and s_ports and origin_s_value:
                center.cli_config_set_all(key, '2000', s_hosts, s_ports)
        center.create_cluster(yes)
        if origin_m_value:
            # cli config restore cluster-node-timeout
            logger.debug('restore cluster node time out')
            center.cli_config_set_all(key, origin_m_value, m_hosts, m_ports)
            if s_hosts and s_ports and origin_s_value:
                v = origin_s_value
                center.cli_config_set_all(key, v, s_hosts, s_ports)

    def clean(self, logs=False):
        """Clean cluster
        """
        if not isinstance(logs, bool):
            logger.error("option '--logs' can use only 'True' or 'False'")
            return
        center = Center()
        center.update_ip_port()
        if logs:
            center.remove_all_of_redis_log_force()
            return
        center.cluster_clean()

    def use(self, cluster_id):
        """Change selected cluster

        :param cluster_id: target cluster #
        """
        _change_cluster(cluster_id)
        cluster_id = '-' if cluster_id == -1 else cluster_id
        logger.info("Cluster '{}' selected.".format(cluster_id))

    def ls(self):
        """Check cluster list"""
        logger.info(cluster_util.get_cluster_list())

    def restart(
        self,
        force_stop=False,
        reset=False,
        cluster=False,
        profile=False,
        yes=False,
    ):
        """Restart redist cluster
        :param force: If true, send SIGKILL. If not, send SIGINT
        :param reset: If true, clean(rm data).
        """
        if not isinstance(force_stop, bool):
            msg = [
                "option '--force-stop' can use only ",
                "'True' or 'False'",
            ]
            logger.error(''.join(msg))
            return
        if not isinstance(reset, bool):
            logger.error("option '--reset' can use only 'True' or 'False'")
            return
        if not isinstance(cluster, bool):
            logger.error("option '--cluster' can use only 'True' or 'False'")
            return
        if not reset and cluster:
            msg = "option '--cluster' can used only with option '--reset'"
            logger.error(msg)
            return
        if not cluster and yes:
            msg = "option '--yes' can used only with option '--cluster'"
            logger.error(msg)
            return
        center = Center()
        center.update_ip_port()
        success = center.check_hosts_connection()
        if not success:
            return
        center.stop_redis(force=force_stop)
        if reset:
            self.clean()
        self.start(profile=profile)
        if cluster:
            self.create(yes=yes)

    def configure(self):
        center = Center()
        center.update_ip_port()
        success = center.check_hosts_connection()
        if not success:
            return
        center.configure_redis()
        center.sync_conf(show_result=True)

    def rowcount(self):
        """Query and show cluster row count"""
        logger.debug('rowcount')
        # open-redis-cli-all info Tablespace | grep totalRows | awk -F ',
        # ' '{print $4}' | awk -F '=' '{sum += $2} END {print sum}'
        host_list = config.get_master_host_list()
        port_list = config.get_master_port_list()
        outs, _ = RedisCliUtil.command_raw_all(
            'info Tablespace', host_list, port_list)
        lines = outs.splitlines()
        key = 'totalRows'
        filtered_lines = (filter(lambda x: key in x, lines))
        ld = RedisCliUtil.to_list_of_dict(filtered_lines)
        # row_count = reduce(lambda x, y: {key: int(x[key]) + int(y[key])}, ld)
        row_count = reduce(lambda x, y: x + int(y[key]), ld, 0)
        self._print(row_count)

    def rebalance(self, ip, port):
        """Rebalance

        :param ip: rebalance target ip
        :param port: rebalance target port
        """
        rebalance_cluster_cmd(ip, port)

    def add_slave(self, yes=False):
        """Add slaves to cluster additionally
        """
        logger.debug('add_slave')
        if not isinstance(yes, bool):
            logger.error("option '--yes' can use only 'True' or 'False'")
            return
        center = Center()
        center.update_ip_port()
        # check
        s_hosts = center.slave_host_list
        s_ports = center.slave_port_list
        if not s_hosts:
            raise ClusterRedisError('Slave host cannot empty')
        if not s_ports:
            raise ClusterRedisError('Slave port cannot empty')
        success = center.check_hosts_connection(hosts=s_hosts)
        if not success:
            return
        center.ensure_cluster_exist()
        slave_alive_count = center.get_alive_slave_redis_count()
        if slave_alive_count > 0:
            msg = [
                'Fail to start slave nodes... ',
                'Must be checked running slave processes!\n',
                'We estimate that ',
                "redis 'SLAVE' processes is {}".format(slave_alive_count)
            ]
            raise LightningDBError(12, ''.join(msg))

        # confirm info
        result = center.confirm_node_port_info(skip=yes)
        if not result:
            logger.warn('Cancel add-slave')
            return
        # clean
        center.cluster_clean(master=False)
        # backup logs
        center.backup_server_logs(master=False)
        center.create_redis_data_directory(master=False)
        # configure
        center.configure_redis(master=False)
        center.sync_conf()
        # start
        center.start_redis_process(master=False)
        center.wait_until_all_redis_process_up()

        # change redis config temporarily
        key = 'cluster-node-timeout'
        origin_s_value = center.cli_config_get(key, s_hosts[0], s_ports[0])
        if not origin_s_value:
            msg = "RedisConfigKeyError: '{}'".format(key)
            logger.warning(msg)
        if origin_s_value:
            # cli config set cluster-node-timeout 2000
            logger.debug('set cluster node time out 2000 for create')
            center.cli_config_set_all(key, '2000', s_hosts, s_ports)
        # create
        center.replicate()
        if origin_s_value:
            # cli config restore cluster-node-timeout
            logger.debug('restore cluster node time out')
            center.cli_config_set_all(key, origin_s_value, s_hosts, s_ports)

    def failover(self):
        """Replace disconnected master with slave.

        Replace disconnected master with slave.
        If disconnected master comes back to live, it become slave.
        """
        center = Center()
        center.update_ip_port()
        master_obj_list = center.get_master_obj_list()
        msg = color.yellow('{} has no alive slave to proceed failover')
        all_alive = True
        for node in master_obj_list:
            if node['status'] != 'connected':
                all_alive = False
                success = False
                for slave in node['slaves']:
                    if slave['status'] == 'connected':
                        logger.info('failover {} for {}'.format(
                            slave['addr'],
                            node['addr']
                        ))
                        stdout = center.run_failover(
                            slave['addr'],
                            take_over=True
                        )
                        if stdout != 'OK':
                            continue
                        logger.info('OK')
                        success = True
                        break
                if not success:
                    logger.info(msg.format(node['addr']))
        if all_alive:
            logger.info('All master is alive')

    def failback(self):
        center = Center()
        center.update_ip_port()
        master_obj_list = center.get_master_obj_list()
        disconnected_list = []
        paused_list = []
        for master in master_obj_list:
            if master['status'] == 'disconnected':
                disconnected_list.append(master['addr'])
            if master['status'] == 'paused':
                paused_list.append(master['addr'])
            for slave in master['slaves']:
                if slave['status'] == 'disconnected':
                    disconnected_list.append(slave['addr'])
                if slave['status'] == 'paused':
                    paused_list.append(slave['addr'])
        classified_disconnected_list = {}
        classified_paused_list = {}
        for disconnected in disconnected_list:
            host, port = disconnected.split(':')
            if host not in classified_disconnected_list:
                classified_disconnected_list[host] = []
            classified_disconnected_list[host].append(port)
        for paused in paused_list:
            host, port = paused.split(':')
            if host not in classified_paused_list:
                classified_paused_list[host] = []
            classified_paused_list[host].append(port)
        current_time = time.strftime("%Y%m%d-%H%M", time.gmtime())
        for host, ports in classified_disconnected_list.items():
            logger.info('run {}:{}'.format(host, '|'.join(ports)))
            center.run_redis_process(host, ports, False, current_time)
        for host, ports in classified_paused_list.items():
            logger.info('restart {}:{}'.format(host, '|'.join(ports)))
            center.stop_redis_process(host, ports)
            center.run_redis_process(host, ports, False, current_time)
        if not classified_disconnected_list and not classified_paused_list:
            logger.info('All redis is alive')

    def tree(self):
        """The results of 'cli cluster nodes' are displayed in tree format.
        """
        center = Center()
        center.update_ip_port()
        master_node_list = center.get_master_obj_list()
        output_msg = []
        for master_node in master_node_list:
            addr = master_node['addr']
            status = master_node['status']
            msg = '{}({})'.format(addr, status)
            if status == 'disconnected':
                msg = color.red(msg)
            if status == 'paused':
                msg = color.yellow(msg)
            output_msg.append(msg)
            for slave_node in master_node['slaves']:
                addr = slave_node['addr']
                status = slave_node['status']
                msg = '{}({})'.format(addr, status)
                if status == 'disconnected':
                    msg = color.red(msg)
                if status == 'paused':
                    msg = color.yellow(msg)
                output_msg.append('|__ ' + msg)
            output_msg.append('')
        logger.info(color.ENDC + '\n'.join(output_msg))

    def _print(self, text):
        if self._print_mode == 'screen':
            logger.info(text)



class LtcliBaseError(Exception):
    '''base error for ltcli'''
    def __init__(self, message, *args):
        Exception.__init__(self, message, *args)
        self.message = message

    def __str__(self):
        return self.message

    def class_name(self):
        '''return class name'''
        return self.__class__.__name__


class LightningDBError(LtcliBaseError):
    def __init__(self, error_code, message, *args):
        self.error_code = error_code
        LtcliBaseError.__init__(self, message, *args)


class ConvertError(LtcliBaseError):
    def __init__(self, message, *args):
        LtcliBaseError.__init__(self, message, *args)


class PropsKeyError(LtcliBaseError):
    def __init__(self, key, *args):
        message = "{} cannot empty in props".format(key.upper())
        LtcliBaseError.__init__(self, message, *args)


class PropsError(LtcliBaseError):
    def __init__(self, message, *args):
        LtcliBaseError.__init__(self, message, *args)


class FileNotExistError(LtcliBaseError):
    def __init__(self, file_path, **kwargs):
        message = "'{}'".format(file_path)
        if 'host' in kwargs.keys():
            self.host = kwargs['host']
            message = "'{}' at '{}'".format(file_path, self.host)
        LtcliBaseError.__init__(self, message, *kwargs)


class SSHConnectionError(LtcliBaseError):
    def __init__(self, host, *args):
        message = "SSH connection fail to '{}'".format(host)
        LtcliBaseError.__init__(self, message, *args)


class HostConnectionError(LtcliBaseError):
    def __init__(self, host, *args):
        message = "Host connection fail to '{}'".format(host)
        LtcliBaseError.__init__(self, message, *args)


class HostNameError(LtcliBaseError):
    def __init__(self, host, *args):
        message = "Unknown host name '{}'".format(host)
        LtcliBaseError.__init__(self, message, *args)


class YamlSyntaxError(LtcliBaseError):
    def __init__(self, file_path, *args):
        LtcliBaseError.__init__(self, "'{}'".format(file_path), *args)


class PropsSyntaxError(LtcliBaseError):
    def __init__(self, line, line_number, *args):
        message = "'{}' at line {}".format(line, line_number)
        LtcliBaseError.__init__(self, message, *args)


class ClusterIdError(LtcliBaseError):
    def __init__(self, cluster_id, *args):
        message = "Invalid cluster id '{}'".format(cluster_id)
        LtcliBaseError.__init__(self, message, *args)


class ClusterNotExistError(LtcliBaseError):
    def __init__(self, cluster_id, **kwargs):
        message = "Not exist cluster '{}'".format(cluster_id)
        if 'host' in kwargs.keys():
            self.host = kwargs['host']
            message = "{} at '{}'".format(message, self.host)
        LtcliBaseError.__init__(self, message, *kwargs)


class ClusterRedisError(LtcliBaseError):
    def __init__(self, message, *args):
        LtcliBaseError.__init__(self, message, *args)


class SSHCommandError(LtcliBaseError):
    def __init__(self, exit_status, host, stderr, *args):
        self.exit_status = exit_status
        self.host = host
        self.stderr = stderr
        message = "[ExitCode {}] Fail execute command at '{}': {}".format(
            exit_status,
            host,
            stderr
        )
        LtcliBaseError.__init__(self, message, *args)


class CreateDirError(LtcliBaseError):
    def __init__(self, message, dir_path, *args):
        self.dir_path = dir_path
        LtcliBaseError.__init__(self, message, *args)


class EnvError(LtcliBaseError):
    def __init__(self, env, *args):
        message = 'you should set env {}'.format(env)
        LtcliBaseError.__init__(self, message, *args)



from ltcli import utils
from ltcli.log import logger
from ltcli import color
from ltcli.rediscli import (
    RedisCliCluster,
    RedisCliConfig,
    RedisCliInfo,
    RedisCliUtil
)


# pylint: disable=redefined-builtin
# need to parameter all for cli options
def _command(sub_cmd, all, host, port):
    if all:
        RedisCliUtil.command_all(sub_cmd=sub_cmd, formatter=utils.print_table)
    else:
        RedisCliUtil.command(sub_cmd=sub_cmd, host=host, port=port)


def ping(host=None, port=None, all=False):
    """Send ping command

    :param all: If true, send command to all
    :param host: host info
    :param port: port info
    """
    if not isinstance(all, bool):
        logger.error("option '--all' can use only 'True' or 'False'")
        return
    if (not host or not port) and not all:
        logger.error("Enter host and port or use '--all' option.")
        return
    if all:
        meta = []
        ret = RedisCliUtil.command_all_async('ping 2>&1')
        pong_cnt = 0
        for m_s, host, port, result, _ in ret:
            addr = '{}:{}'.format(host, port)
            if result == 'OK':
                pong_cnt += 1
            else:
                meta.append([m_s, addr, color.red('FAIL')])
        if meta:
            utils.print_table([['TYPE', 'ADDR', 'RESULT']] + meta)
        logger.info('alive redis {}/{}'.format(pong_cnt, len(ret)))
        return
    if host and port:
        _command('ping', False, host, port)


def reset_oom(all=False, host=None, port=0):
    """Send reset oom command

    :param all: If true, send command to all
    :param host: host info
    :param port: port info
    """
    if not isinstance(all, bool):
        logger.error("option '--all' can use only 'True' or 'False'")
        return
    sub_cmd = 'resetOom'
    _command(sub_cmd, all, host, port)


def reset_info(key, all=False, host=None, port=0):
    """Send reset info

    :param key: resetting target key string
    :param all: If true, send command to all
    :param host: host info
    :param port: port info
    """
    if not isinstance(all, bool):
        logger.error("option '--all' can use only 'True' or 'False'")
        return
    sub_cmd = 'resetInfo %s' % key
    _command(sub_cmd, all, host, port)


def metakeys(key, all=False, host=None, port=0):
    """Get metakeys

    :param key: resetting target key string
    :param all: If true, send command to all
    :param host: host info
    :param port: port info
    """
    if not isinstance(all, bool):
        logger.error("option '--all' can use only 'True' or 'False'")
        return
    sub_cmd = 'metakeys "%s"' % key
    _command(sub_cmd, all, host, port)


class Cli(object):
    """This is Cli command (redis-cli wrapper)

    You can check redis and cluster info.

    """

    def __init__(self):
        self.info = RedisCliInfo()
        self.config = RedisCliConfig()
        self.cluster = RedisCliCluster()
        self.ping = ping
        self.reset_oom = reset_oom
        self.reset_info = reset_info
        self.metakeys = metakeys
